#!/usr/bin/env python
#
# Scanner for a calculator interpreter

import cStringIO as StringIO
import string

import calc_scanner


class Parser(object):
    """Implement a Parser for the following grammar:
    
       Code      :== Expr (',' EXPR)* EOT
       Expr      :== SecExpr (AddOp SecExpr)*
       SecExpr   :== PriExpr (MultOp PriExpr)*
       PriExpr   :== INT | FLOAT | STR | IDENTSUB | '(' EXPR ')' |
                     IDENT LPAREN EXPR (',' EXPR)* RPAREN | '-' EXPR
       AddOp     :== '+' | '-' | '||' 
       MulOp     :== '*' | '/' | '%'
    """
    def __init__(self, tokens):
        self.tokens = tokens
        self.curindex = 0
        self.curtoken = tokens[0]
    
    def parse(self):
        ast = self.parseCode()
        return ast
        
    def parseCode(self):
        e1AST = self.parseExpr()
        exprs = [e1AST]
        token = self.tokenCurrent()
        while (token.type == TK_COMMA):
            self.tokenAcceptAny()
            e1AST = self.parseExpr()
            exprs.append(e1AST)
            token = self.tokenCurrent()
        self.tokenAccept(TK_EOT)
        return ExprCode(exprs)
    
    def parseExpr(self):
        e1AST = self.parseSecExpr()
        token = self.tokenCurrent()
        while (token.type == TK_ADD or token.type == TK_SUB or token.type == TK_CONCAT):
            op = token
            self.tokenAcceptAny()
            e2AST = self.parseSecExpr()
            e1AST = ExprBinary(e1AST, op, e2AST)
            token = self.tokenCurrent()
        return e1AST
    
    def parseSecExpr(self):
        e1AST = self.parsePriExpr()
        token = self.tokenCurrent()
        while (token.type == TK_MUL or token.type == TK_DIV or token.type == TK_MOD):
            op = token
            self.tokenAcceptAny()
            e2AST = self.parsePriExpr()
            e1AST = ExprBinary(e1AST, op, e2AST)
            token = self.tokenCurrent()
        return e1AST
    
    def parsePriExpr(self):
        token = self.tokenCurrent()
        if token.type == TK_INT:
            self.tokenAcceptAny()
            return ExprValue(token)
        elif token.type == TK_FLOAT:
            self.tokenAcceptAny()
            return ExprValue(token)
        elif token.type == TK_STR:
            self.tokenAcceptAny()
            return ExprValue(token)
        elif token.type == TK_IDENTSUB:
            self.tokenAcceptAny()
            return ExprVar(token)
        elif token.type == TK_LPAREN:
            self.tokenAcceptAny()
            e1AST = self.parseExpr()
            self.tokenAccept(TK_RPAREN)
            return e1AST
        elif token.type == TK_IDENT:
            func = token
            self.tokenAcceptAny()
            self.tokenAccept(TK_LPAREN)
            e1AST = self.parseExpr()
            args = [e1AST]
            token = self.tokenCurrent()
            while (token.type == TK_COMMA):
                self.tokenAcceptAny()
                e1AST = self.parseExpr()
                args.append(e1AST)
                token = self.tokenCurrent()
            self.tokenAccept(TK_RPAREN)
            return ExprCall(func, args)
        elif token.type == TK_SUB:
            self.tokenAcceptAny()
            e1AST = self.parseExpr()
            return ExprUnary(token, e1AST)
        else:
            raise ParserError(token.pos)

    def tokenCurrent(self):
        return self.curtoken
        
    def tokenAcceptAny(self):
        if self.curtoken.type != TK_EOT:
            self.curindex += 1
            self.curtoken = self.tokens[self.curindex]

    def tokenAccept(self, type):
        if self.curtoken.type != type:
            raise ParserError(self.curtoken.pos)
        self.tokenAcceptAny()
